#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <string.h>        /* memcpy */

#include <stdio.h>
#include <stdlib.h>

#include "color_correct.h"
#include "data.h"
#include "colorconv.h"

extern GimpDrawable *src_drawable_global;

static gchar *levels_bg = NULL;

gboolean
on_levels_drawingarea_button_press_event (GtkWidget      *widget,
                                          GdkEventButton *event,
                                          gpointer        user_data);

gboolean
on_levels_drawingarea_configure_event (GtkWidget         *widget,
                                       GdkEventConfigure *event,
                                       gpointer           user_data);

gboolean
on_levels_drawingarea_motion_notify_event (GtkWidget      *widget,
                                           GdkEventMotion *event,
                                           gpointer        user_data);

gboolean
on_levels_drawingarea_expose_event (GtkWidget      *widget,
                                    GdkEventExpose *event,
                                    gpointer        user_data);

GtkWidget *by;
GtkWidget *gy;
GtkWidget *wy;
GtkWidget *outb;
GtkWidget *outw;
GtkWidget *levels_drawingarea;


GtkWidget *
levels_ui (void)
{
  GtkWidget *levels_vbox;
  GtkWidget *levels_frame;
  GtkWidget *levels_values;
  GtkWidget *input_label;
  GtkWidget *output_label;
  GtkObject *by_adj;
  GtkObject *gy_adj;
  GtkObject *wy_adj;
  GtkObject *outb_adj;
  GtkObject *outw_adj;

  levels_vbox = gtk_vbox_new (FALSE, 0);

  gtk_container_set_border_width (GTK_CONTAINER (levels_vbox), 2);

  levels_frame = gtk_frame_new (_("input and output levels"));
  gtk_box_pack_start (GTK_BOX (levels_vbox), levels_frame, FALSE, FALSE, 0);

  levels_drawingarea = gtk_drawing_area_new ();

  gtk_container_add (GTK_CONTAINER (levels_frame), levels_drawingarea);
  gtk_widget_set_size_request (levels_drawingarea, 256, 194);
  gtk_widget_set_events (levels_drawingarea,
               GDK_EXPOSURE_MASK | GDK_BUTTON1_MOTION_MASK |
               GDK_BUTTON_PRESS_MASK);

  levels_values = gtk_table_new (2, 4, FALSE);
  gtk_box_pack_start (GTK_BOX (levels_vbox), levels_values, TRUE, TRUE, 0);

  output_label = gtk_label_new (_("Output values: "));
  gtk_table_attach (GTK_TABLE (levels_values), output_label, 0, 1, 1, 2,
            (GtkAttachOptions) (GTK_FILL),
            (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (output_label), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (output_label), 0, 0.5);

  input_label = gtk_label_new (_("Input values: "));
  gtk_table_attach (GTK_TABLE (levels_values), input_label, 0, 1, 0, 1,
            (GtkAttachOptions) (GTK_FILL),
            (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (input_label), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (input_label), 0, 0.5);

  by_adj = gtk_adjustment_new (0, 0, 255, 1, 10, 10);
  by = gtk_spin_button_new (GTK_ADJUSTMENT (by_adj), 1, 0);
  gtk_table_attach (GTK_TABLE (levels_values), by, 1, 2, 0, 1,
            (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
            (GtkAttachOptions) (0), 0, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (by), TRUE);

  gy_adj = gtk_adjustment_new (128, 0, 255, 1, 10, 10);
  gy = gtk_spin_button_new (GTK_ADJUSTMENT (gy_adj), 1, 0);
  gtk_table_attach (GTK_TABLE (levels_values), gy, 2, 3, 0, 1,
            (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
            (GtkAttachOptions) (0), 0, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (gy), TRUE);

  wy_adj = gtk_adjustment_new (255, 0, 255, 1, 10, 10);
  wy = gtk_spin_button_new (GTK_ADJUSTMENT (wy_adj), 1, 0);
  gtk_table_attach (GTK_TABLE (levels_values), wy, 3, 4, 0, 1,
            (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
            (GtkAttachOptions) (0), 0, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (wy), TRUE);

  outb_adj = gtk_adjustment_new (0, 0, 255, 1, 10, 10);
  outb = gtk_spin_button_new (GTK_ADJUSTMENT (outb_adj), 1, 0);
  gtk_table_attach (GTK_TABLE (levels_values), outb, 1, 2, 1, 2,
            (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
            (GtkAttachOptions) (0), 0, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (outb), TRUE);

  outw_adj = gtk_adjustment_new (255, 0, 255, 1, 2, 2);
  outw = gtk_spin_button_new (GTK_ADJUSTMENT (outw_adj), 1, 0);
  gtk_table_attach (GTK_TABLE (levels_values), outw, 3, 4, 1, 2,
            (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
            (GtkAttachOptions) (0), 0, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (outw), TRUE);

  g_signal_connect (GTK_OBJECT (levels_drawingarea), "button_press_event",
            GTK_SIGNAL_FUNC
            (on_levels_drawingarea_button_press_event), NULL);
  g_signal_connect (GTK_OBJECT (levels_drawingarea), "configure_event",
            GTK_SIGNAL_FUNC (on_levels_drawingarea_configure_event),
            NULL);
  g_signal_connect (GTK_OBJECT (levels_drawingarea), "motion_notify_event",
            GTK_SIGNAL_FUNC
            (on_levels_drawingarea_motion_notify_event), NULL);
  g_signal_connect (GTK_OBJECT (levels_drawingarea), "expose_event",
            GTK_SIGNAL_FUNC (on_levels_drawingarea_expose_event),
            NULL);
  g_signal_connect (GTK_OBJECT (by), "changed",
            GTK_SIGNAL_FUNC (on_values_changed), NULL);
  g_signal_connect (GTK_OBJECT (gy), "changed",
            GTK_SIGNAL_FUNC (on_values_changed), NULL);
  g_signal_connect (GTK_OBJECT (wy), "changed",
            GTK_SIGNAL_FUNC (on_values_changed), NULL);
  g_signal_connect (GTK_OBJECT (outb), "changed",
            GTK_SIGNAL_FUNC (on_values_changed), NULL);
  g_signal_connect (GTK_OBJECT (outw), "changed",
            GTK_SIGNAL_FUNC (on_values_changed), NULL);

  return levels_vbox;
}



gboolean
on_levels_drawingarea_button_press_event (GtkWidget      *widget,
                                          GdkEventButton *event,
                                          gpointer        user_data)
{
  gint xorigin = widget->allocation.width / 2 - (256 / 2);
  gint yorigin = widget->allocation.height / 2 - (146 / 2);

  gint x, y;

  x = (event->x) - xorigin;
  y = (event->y) - yorigin;

  if (x < 0)
    x = 0;
  if (x > 255)
    x = 255;

  if (y >= 130) {
    if (x < pv.outb) {
      pv.outb = x;
    } else if (x > pv.outw) {
      pv.outw = x;
    } else {
      if (pv.outw - x < x - pv.outb) {
        pv.outw = x;
      } else {
        pv.outb = x;
      }
    }
  } else {
    if (x < pv.by) {
      pv.by = x;
    } else if (x > pv.wy) {
      pv.wy = x;
    } else if (x < pv.gy) {
      if (pv.gy - x < x - pv.by) {
        pv.gy = x;
      } else {
        pv.by = x;
      }
    } else {        /* x>gy ) */
      if (pv.wy - x < x - pv.gy) {
        pv.wy = x;
      } else {
        pv.gy = x;
      }
    }
  }

  update_values ();
  gtk_widget_queue_draw (widget);


  return FALSE;
}



gboolean
on_levels_drawingarea_configure_event (GtkWidget         *widget,
                                       GdkEventConfigure *event,
                                       gpointer           user_data)
{
  if (!levels_bg) {
    guint histogram_data[256];
    gint x, y;
    gchar *d;

    levels_bg = g_malloc0 (256 * 146 * 3);

    /* draw grayscale at bottom */
    for (x = 0; x < 256; x++)
      for (y = 130; y < 146; y++) {
        d = &levels_bg[(y * 256 * 3) + x * 3];
        d[0] = x;
        d[1] = x;
        d[2] = x;
      }

    memset (histogram_data, 0, sizeof (histogram_data));

    {            /*FIXME: only collect from selection area */
      GimpPixelRgn rgn;
      gpointer pr;

      gimp_pixel_rgn_init (&rgn, src_drawable_global, 0, 0,
                 src_drawable_global->width,
                 src_drawable_global->height, FALSE, FALSE);
      for (pr = gimp_pixel_rgns_register (1, &rgn); pr != NULL;
         pr = gimp_pixel_rgns_process (pr)) {
        int x, y;
        guchar *src = rgn.data;

        for (y = 0; y < rgn.h; y++) {
          guchar *s = src;

          for (x = 0; x < rgn.w; x++) {
            int R, G, B, val;

            R = s[0];
            G = s[1];
            B = s[2];
            val = (R + G + B) / 3;

            histogram_data[val]++;

            s += rgn.bpp;
          }
          src += rgn.rowstride;
        }
      }
    }

    {
      int pos;
      int max = 0;

      for (pos = 2; pos < 254; pos++)
        if (histogram_data[pos] > max)
          max = histogram_data[pos];
      for (pos = 0; pos < 256; pos++)
        histogram_data[pos] = (histogram_data[pos] * 256) / max;
    }


    for (x = 0; x < 256; x++) {
      int i = histogram_data[x];

      i /= 2;
      for (y = 0; y < 128; y++) {
        d = &levels_bg[(y * 256 * 3) + x * 3];

        if (128 - y > i) {
          d[0] = 244;
          d[1] = 244;
          d[2] = 244;
        } else {
          d[0] = 140;
          d[1] = 140;
          d[2] = 140;
        }
      }
    }

  }
  return FALSE;
}


gboolean
on_levels_drawingarea_motion_notify_event (GtkWidget      *widget,
                                           GdkEventMotion *event,
                                           gpointer        user_data)
{
  return on_levels_drawingarea_button_press_event (widget,
                           (GdkEventButton *) event,
                           user_data);
}


gboolean
on_levels_drawingarea_expose_event (GtkWidget      *widget,
                                    GdkEventExpose *event,
                                    gpointer        user_data)
{
  int xorigin = widget->allocation.width / 2 - (256 / 2);
  int yorigin = widget->allocation.height / 2 - (146 / 2);

  gdk_draw_rgb_image (widget->window, widget->style->white_gc,
            xorigin, yorigin,
            256, 146, GDK_RGB_DITHER_NONE, levels_bg, 256 * 3);

  gdk_draw_rectangle (widget->window, widget->style->black_gc, TRUE,
            xorigin + pv.by, yorigin, 1, 128);

  gdk_draw_rectangle (widget->window, widget->style->black_gc, TRUE,
            xorigin + pv.gy, yorigin, 1, 128);

  gdk_draw_rectangle (widget->window, widget->style->black_gc, TRUE,
            xorigin + pv.wy, yorigin, 1, 128);

  gdk_draw_rectangle (widget->window, widget->style->white_gc, TRUE,
            xorigin + pv.outb, yorigin + 130, 1, 16);

  gdk_draw_rectangle (widget->window, widget->style->black_gc, TRUE,
            xorigin + pv.outw, yorigin + 130, 1, 16);

  return FALSE;
}
